# -*- coding: utf-8 -*-

from odoo import api, fields, models
from odoo.addons.saas_worker.const import TEST, WEBSITE

from hashlib import sha256
from dateutil.relativedelta import relativedelta
from werkzeug.urls import url_encode

import base64
import hmac
import logging
import pprint
import requests
import time

_logger = logging.getLogger(__name__)


# US shipping SKU: Odoo-Mug-White-15oz-US
# WorldWide shipping from US except US: Odoo-Mug-White-15oz-WorldWide
def _gooten_order_sku(sku, recipeId, billingKey, quantity=1, **kwargs):
    gooten_data = {
        'IsInTestMode': TEST,
        'ShipToAddress': {
            "FirstName": kwargs["firstName"],
            "LastName": kwargs["lastName"],
            "Line1": kwargs["addrLine1"],
            "Line2": kwargs.get("addrLine2"),
            "City": kwargs["addrCity"],
            "PostalCode": kwargs["addrPostalCode"],
            "CountryCode": kwargs["addrCountryCode"],
            "Phone": kwargs["phone"],
            "Email": kwargs["email"],
            "IsBusinessAddress": False,
        },
        'BillingAddress': {
            "FirstName": "Odoo S.A.",
            "LastName": "Odoo S.A.",
            "Line1": "40 chaussée de Namur",
            "City": "Ramillies",
            "CountryCode": "BE",
            "PostalCode": "1367",
            "IsBusinessAddress": True,
            "Phone": "003281813700",
            "Email": "sys@odoo.com",
        },
        'Items': [
            {
                'Quantity': quantity,
                'SKU': sku,
                'ShipCarrierMethodId': 0,
            }
        ],
        'Payment': {
            'PartnerBillingKey': billingKey,
        }
    }

    if kwargs.get("addrState"):
        gooten_data['ShipToAddress']['State'] = kwargs['addrState']

    gooten_api_url = 'https://api.print.io/api/v/5/source/api/orders/'
    try:
        response = requests.post(gooten_api_url, json=gooten_data, params={'recipeid': recipeId})
        response = response.json()
    except Exception as e:
        gooten_data.pop('Payment', None) # this is secret, we do not want it logged
        _logger.error("_gooten_order_sku: exception when ordering on gooten, msg %s, sent values (redacted)\n%s", e, pprint.pformat(gooten_data))
        return False

    if response.get('HadError'):
        _logger.error("_gooten_order_sku: Unable to order sku %s (%d), received values\n%s", sku, quantity, pprint.pformat(response))
        return False

    _logger.info("_gooten_order_sku: ordered SKU \"%s\", order id \"%s\"", sku, response['Id'])
    return response['Id']



class MugToken(models.Model):
    _name = 'openerp.enterprise.mug.token'
    _description = 'List of token for Odoo Mug redeem'
    _rec_name = 'token'

    token = fields.Char('Token', required=True)

    @api.autovacuum
    def _gc_tokens(self):
        self.env.cr.execute("""
            DELETE FROM %s
                WHERE
                    create_date < %%s
            """ % self._table,
            (fields.Datetime.to_string(fields.Datetime.now() - relativedelta(days=7)),)
        )


class OpenerpEnterpriseDatabaseUser(models.Model):

    _inherit = "openerp.enterprise.database.user"


    # SaaS invitation gift program
    gift_invited_mails = fields.Char('Invited mails') # comma separated mail addresses
    gift_gooten_order_id = fields.Char('Gooten Order Id', default='')
    gift_invitation_state = fields.Selection([
        ('none', 'Not proposed'),           # we have not proposed the gift for the user
        ('skipped', 'Skipped'),             # The user didn't invite anyone or invited less than the required amount for the gift
        ('pending', 'Pending invites'),     # The user invited people and is proposed the gift
        ('done', 'Users registered')        # The invited users registered
    ], default='none', string='Gift invitation state')
    gift_redeemed = fields.Boolean('Gift redeemed', compute='_compute_gift_redeemed', readonly=True)
    gift_security_token = fields.Char('Gift security token', compute='_compute_gift_security_token', groups="openerp_enterprise.group_manager")
    connected_once = fields.Boolean('Has connected once', compute='_compute_connected_once', search='_search_connected_once')
    invite_count = fields.Integer(string='Invited User count', compute='_compute_invite_count')
    registered_count = fields.Integer(string='Invited users who registered', compute='_compute_registered_count')

    @api.depends('gift_invited_mails')
    def _compute_invite_count(self):
        for r in self:
            if r.gift_invited_mails:
                r.invite_count = len(r.gift_invited_mails.split(','))
            else:
                r.invite_count = 0

    @api.depends('gift_invited_mails')
    def _compute_registered_count(self):
        for r in self:
            if r.gift_invited_mails and len(r.gift_invited_mails):
                mails = r.gift_invited_mails.split(',')
                r.registered_count = self.search_count([('database_id', '=', r.database_id.id),
                                                    ('login', 'in', mails),
                                                    ('login_date', '!=', False)])
            else:
                r.registered_count = 0

    def _compute_gift_redeemed(self):
        for r in self:
            r.gift_redeemed = bool(r.gift_gooten_order_id)

    def _compute_connected_once(self):
        for r in self:
            # if the user has set a password set, we consider it has connected once
            r.connected_once = bool(r.login_date)

    def _search_connected_once(self, operator, value):
        if (operator, value) in [('=', True), ('!=', False)]:
            return [('login_date', '!=', False)]
        return [('login_date', '=', False)]

    def _compute_gift_security_token(self):
        db_secret = self.env['ir.config_parameter'].sudo().get_param('database.secret').encode()
        for r in self:
            r.gift_security_token = hmac.new(db_secret, r.login.encode(), sha256).hexdigest()

    def _cron_check_gift_invitation(self):
        if not self:
            # search for all invitation that have not been processed yet
            self = self.search([
                ('gift_invitation_state', '!=', 'done'),
                ('gift_invited_mails', '!=', False),
                '|',
                ('gift_gooten_order_id', '!=', ''),     # not empty
                ('gift_gooten_order_id', '!=', False)   # not null
            ])

        for r in self:
            invited_mails = r.gift_invited_mails.split(',')
            invited_count = self.search_count([('login', 'in', invited_mails), ('connected_once', '=', True), ('database_id', '=', r.database_id.id)])
            # if at least 2 of the invitee connected, then we send the gift
            if invited_count >= 2:
                # all the user that have been invited are now created and logged once.
                if r.gift_invitation_state == 'pending': # otherwise if the host was proposed the gift
                    # create the redeem url with a sha256 of the customer login
                    redeem_gift_url = WEBSITE + '/odoo-enterprise/redeem-invitation-gift?' + url_encode({'uid': r.id, 'h': r.gift_security_token})
                    # send the mail to the "inviter"
                    mail_template = self.env.ref('openerp_enterprise.oe_invitation_accepted')
                    mail_template.with_context({
                        'invited_mails_count': invited_count,
                        'redeem_url': redeem_gift_url,
                    }).send_mail(r.id)
                    self.env.cr.commit()
                r.gift_invitation_state = 'done'


    @api.model
    def _gooten_order_mug(self, **kwargs):
        """Method to order the Odoo mug at Gooten.
            Required values are stored in kwargs (too much of them)
            param str firstName: Receiver's first name (required).
            param str lastName: Receiver's last name (required).
            param str addrLine1: Receiver's address (required).
            param str addrLine2: Receiver's address (optional).
            param str addrCity: Receiver's city (required).
            param str addrState: Receiver's State (optional).
            param str addrPostalCode: Receiver's postal code (required).
            param str addrCountryCode: Receiver's country code (required).
            param str phone: Receiver's phone number (required).
            param str email: Receiver's email address (required)."""
        # gooten has two way of delivering which are really confusing.
        # They ship from US to worldwide but no directly to US
        # To ship to US you have a order a product made specificaly for US & Canada
        SKU = "Mug-White-11oz-20210728135653989"
        if kwargs['addrCountryCode'] in ['US']:
            SKU = "Odoo-Mug-White-15oz-US"

        # order one odoo mug
        gooten_recipe_id = self.env['ir.config_parameter'].sudo().get_param('saas_invite.token_gooten_recipe_id')
        gooten_billing_key = self.env['ir.config_parameter'].sudo().get_param('saas_invite.token_gooten_billing_key')
        gooten_order_id = _gooten_order_sku(sku=SKU,
                                recipeId=gooten_recipe_id,
                                billingKey=gooten_billing_key,
                                firstName=kwargs['firstName'],
                                lastName=kwargs['lastName'],
                                addrLine1=kwargs['addrLine1'],
                                addrLine2=kwargs.get('addrLine2'),
                                addrCity=kwargs['addrCity'],
                                addrState=kwargs.get('addrState'),
                                addrPostalCode=kwargs['addrPostalCode'],
                                addrCountryCode=kwargs['addrCountryCode'],
                                phone=kwargs['phone'],
                                email=kwargs['email'])
        return gooten_order_id

    def _order_gift(self, **kwargs):
        gooten_order_id = self._gooten_order_mug(**kwargs)
        if gooten_order_id is False:
            return False

        self.gift_gooten_order_id = gooten_order_id
        # send the email to track the gift
        mail_template = self.env.ref('openerp_enterprise.oe_invitation_gift_track')
        mail_template.sudo().with_context({
            'track_url': WEBSITE + '/odoo-enterprise/track-order?' + url_encode({'orderid': gooten_order_id})
        }).send_mail(self.id)
        return True


class Users(models.Model):

    _inherit = 'res.users'

    def _get_mug_token(self, expiration=7*24*60*60):
        """ Return a token to order mugs from the
            /odoo-mug/redeem page
        """
        self.ensure_one()
        db_secret = 'Here, have a mug "%s"' % self.env['ir.config_parameter'].sudo().get_param('database.secret')
        h = hmac.new(db_secret.encode(), digestmod=sha256)
        max_ts = int(time.time()) + expiration
        message = '>%s|%s<' % (max_ts, self.id)
        h.update(message.encode())
        return '%s.%s' % (max_ts, base64.b64encode(h.digest()).decode())

    def _check_mug_token(self, token):
        self.ensure_one()
        max_ts, signature = token.split('.')
        db_secret = 'Here, have a mug "%s"' % self.env['ir.config_parameter'].sudo().get_param('database.secret')
        h = hmac.new(db_secret.encode(), digestmod=sha256)
        message = '>%s|%s<' % (max_ts, self.id)
        h.update(message.encode())
        token_redeemed = self.env['openerp.enterprise.mug.token'].sudo().search_count([('token', '=', token)]) > 0
        return hmac.compare_digest(base64.b64decode(signature), h.digest()) and int(time.time()) < int(max_ts) and not token_redeemed

    def get_mug_redeem_url(self):
        self.ensure_one()
        # we want this method to be public so it can be used on mail.template
        # but we want to limit it access to users who have write access to this res.users
        # so the users can only access their own redeem URL
        self.check_access_rights('write')
        self.check_access_rule('write')
        return WEBSITE + '/odoo-mug/redeem?' + url_encode({'uid': self.id, 'token': self._get_mug_token()})
